#define _GNU_SOURCE

/*
	ted: tiny [text] editor
	Copyright (C) 2009 Edward Cree <techsupport.alphasoft@googlemail.com>

	This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>

#define VERSION_MAJ	0
#define VERSION_MIN	2
#define VERSION_REV	4

#define BACKUPS 1 // Default max. backup depth

typedef struct
{
	int nlines;
	char ** buf;
	char * filename;
	bool saved;
	int curline;
	int cline; // old curline - for 'at'.
	int scrline;
	int onerror; // error vector (scripts)
	bool oeabs; // error vector is absolute? (scripts)
	int backup_depth;
	int err; // cmd_ functions return errors in here
}
buffer;

#define VERSION_MSG "ted %hhu.%hhu.%hhu\n\
 Copyright (C) 2009 Edward Cree.\n\
 License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n\
 This is free software: you are free to change and redistribute it.\n\
 There is NO WARRANTY, to the extent permitted by law.\n",VERSION_MAJ, VERSION_MIN, VERSION_REV

#define GPL_MSG "ted Copyright (C) 2009 Edward Cree.\n\
 This program comes with ABSOLUTELY NO WARRANTY; for details type `help gpl-w'.\n\
 This is free software, and you are welcome to redistribute it\n\
 under certain conditions; type `help gpl-c' for details.\n"

#define USAGE_MSG "Usage:\n\
ted [<filename>] [{-s|--script}=<scriptfile>]\n\
ted {--help|--version}\n"

#define max(a,b)	((a)>(b)?(a):(b))
#define min(a,b)	((a)<(b)?(a):(b))

// function prototypes

char * getl(char *);
char * slurp(FILE *);
buffer cmd_new(char *);
buffer cmd_load(char *, bool);
buffer cmd_save(char *, buffer);
int cmd_show(char *, buffer);
int cmd_showpg(buffer);
buffer cmd_move(char *, buffer, bool);
buffer cmd_insert(char *, buffer);
buffer cmd_replace(char *, buffer);
buffer cmd_append(char *, buffer);
buffer cmd_split(char *, buffer);
buffer cmd_join(char *, buffer);
buffer cmd_delete(char *, buffer);
int cmd_copy(char *, buffer);
buffer cmd_paste(char *, buffer);
int cmd_count(char *, buffer);
buffer cmd_close(char *, buffer, bool);
int cmd_help(char *, bool);
char * ruler(int);
int fbackup(char*, int, int);
buffer add_line_to_buffer(buffer, int, char *);
buffer * script_pop(buffer *, int *);
buffer * script_push(buffer *, int *, buffer);

char * paste=NULL; // pastebuffer, globally available (for now)
bool make_backups=true;
unsigned int def_backup_depth=BACKUPS;

int main(int argc, char *argv[])
{
	char * loadfn = NULL;
	char * scriptfn = NULL;
	int arg;
	for (arg=1; arg<argc; arg++)
	{
		if((strcasecmp(argv[arg], "--version") == 0) || (strcasecmp(argv[arg], "-v") == 0))
		{ // print version info and exit
			printf(VERSION_MSG);
			return(0);
		}
		else if((strcasecmp(argv[arg], "--help") == 0) || (strcasecmp(argv[arg], "-?") == 0))
		{ // print invocation help and exit
			printf(USAGE_MSG);
			return(0);
		}
		else if((strncasecmp(argv[arg], "--script=", 9) == 0) || (strncasecmp(argv[arg], "-s=", 3) == 0))
		{ // script file
			char * eqsign=strchr(argv[arg], '=');
			if(eqsign==NULL)
			{
				printf("Internal error (\"impossible\"); -s option not handled\n");
			}
			else
			{
				scriptfn=eqsign+1;
			}
		}
		else if((strcasecmp(argv[arg], "--no-backups") == 0) || (strcasecmp(argv[arg], "-b") == 0))
		{ // don't make backups with [save]
			make_backups=false;
		}
		else if((strncasecmp(argv[arg], "--backup-depth=", 15) == 0) || (strncasecmp(argv[arg], "-d=", 3) == 0))
		{ // depth of backups made with [save]
			char * eqsign=strchr(argv[arg], '=');
			if(eqsign==NULL)
			{
				printf("Internal error (\"impossible\"); -d option not handled\n");
			}
			else
			{
				sscanf(eqsign+1, "%u", &def_backup_depth);
			}
		}
		else
		{ // unrecognised option, assume it's a filename
			loadfn = argv[arg];
		}
	}
	
	printf(GPL_MSG);
	
	buffer * ebuf = (buffer *)malloc(sizeof(buffer)); // main editing buffers
	int nbuffers=0, curbuffer=0;
	ebuf[0].buf = NULL;
	ebuf[0].nlines=0;
	ebuf[0].filename = NULL;
	ebuf[0].saved=true,
	ebuf[0].curline=0;
	ebuf[0].cline=0;
	ebuf[0].scrline=0;
	ebuf[0].onerror=-1;
	ebuf[0].oeabs=true;
	buffer nbuf=ebuf[0]; // null buffer / not-a buffer
	buffer * sbuf = (buffer *)malloc(sizeof(buffer));
	int sbufs=0;
	sbuf[0]=nbuf; // script file buffer
	sbuf[0].filename=strdup("script-buffer: No script running.");
	buffer abuf = cmd_load(".aliases", false); // aliases buffer
	bool cont=true;
	if(loadfn!=NULL)
	{
		printf("%s: ", loadfn);
		int f=open(loadfn, O_CREAT|O_RDWR, S_IFREG|S_IRUSR|S_IWUSR|S_IRGRP);
		close(f);
		ebuf[0]=cmd_load(loadfn, true);
		if(ebuf[0].err!=0)
		{
			return(ebuf[0].err);
		}
		nbuffers++;
	}
	bool scriptmode=false;
	bool userscript=false;
	if(scriptfn!=NULL)
	{
		printf("Script: ");
		sbuf[0]=cmd_load(scriptfn, true);
		if(sbuf[0].err!=0)
		{
			return(sbuf[0].err);
		}
		if(sbuf[0].filename)
			free(sbuf[0].filename);
		sbuf[0].filename = (char *)malloc(strlen(scriptfn)+20);
		sprintf(sbuf[0].filename, "%s [active script]", scriptfn);
		sbufs++;
		scriptmode=true;
	}
	char *more=NULL;
	while(cont)
	{
		bool stop=false;
		buffer * cbuf;
		if(curbuffer==-1)
		{
			cbuf=&abuf;
		}
		else if(curbuffer<=-2)
		{
			int ws=-2-curbuffer;
			if(ws<sbufs)
				cbuf=&sbuf[ws];
		}
		else if(curbuffer<nbuffers)
		{
			cbuf=&ebuf[curbuffer];
		}
		else
		{
			cbuf=&nbuf;
		}
		if((*cbuf).curline>=(*cbuf).nlines)
			(*cbuf).curline=max((*cbuf).nlines-1, 0);
		bool match=false;
		char *cmd;
		if(more)
		{
			cmd=more;
			more=NULL;
		}
		else
		{
			if(scriptmode)
			{
				if(sbufs)
				{
					if(sbuf[0].scrline>=sbuf[0].nlines)
					{
						printf("Script has reached EOF\n");
						cont=false;
						goto screrr;
					}
					else
					{
						cmd=strdup(sbuf[0].buf[sbuf[0].scrline++]);
					}
				}
				else
				{
					scriptmode=false;
					cont=userscript;
					cmd = getl("(ted) ");
				}
			}
			else
			{
				cmd = getl("(ted) ");
			}
		}
		if(cmd==NULL)
		{
			printf("A strange error occurred\n");
			continue;
		}
		char *semi=strchr(cmd, ';');
		if(semi)
		{
			more=strdup(semi+1);
			*semi=0;
		}
		// alias substitution
		{
			int i;
			for(i=0;i<abuf.nlines;i++)
			{
				char * alias=abuf.buf[i];
				char * colon=strchr(alias, ':');
				if(colon!=NULL) // TODO: if it /is/ null, remove_line_from_buffer (which isn't available yet)
				{
					char * aname = strndup(alias, (size_t)(colon-alias));
					char *find=cmd-1;
					while((find=strstr(find+1, aname)))
					{
						char *cmdis=strndup(cmd, find-cmd);
						char * rest = strdup(find+strlen(aname));
						char * ncmd;
						ncmd=(char *)malloc((find-cmd)+strlen(colon)+strlen(rest)+1);
						if(ncmd==NULL)
						{
							printf("A strange error occurred in alias substitution\n");
						}
						else
						{
							free(cmd);
							cmd=ncmd;
							sprintf(cmd, "%s%s%s", cmdis, colon+1, rest);
						}
						free(rest);
					}
					free(aname);
				}
			}
		}
		// on,at handling
		int on,at;
		bool on_set=false, at_set=false;
		buffer *onbuf;
		if(strncmp(cmd, "on ", 3)==0)
		{
			char *sp=strchr(cmd+3, ' ');
			if(sp)
			{
				on_set=true;
				if(cmd[3]=='c')
				{
					sscanf(cmd+4, "%d", &on);
					on=curbuffer+on;
				}
				else if((cmd[3]=='s') && (scriptmode)) // if a script is running, allow the s flag
				{
					sscanf(cmd+4, "%d", &on);
					on-=2; // relative to script-buffer, which is buffer -2
				}
				else
				{
					sscanf(cmd+3, "%d", &on);
					on--; // user-side, it's base-1 addressed
				}
				if(on==-1)
				{
					onbuf=&abuf;
				}
				else if(on<=-2)
				{
					int ws=-2-on;
					if(ws<sbufs)
						onbuf=&sbuf[ws];
					else
						onbuf=&nbuf;
				}
				else if(on<nbuffers)
				{
					onbuf=&ebuf[on];
				}
				else
				{
					onbuf=&nbuf;
				}
				if((*onbuf).curline>=(*onbuf).nlines)
					(*onbuf).curline=max((*onbuf).nlines-1, 0);
				char *ca=strdup(cmd);
				strcpy(cmd, ca+(sp-cmd)+1);
				free(ca);
			}
		}
		if(strncmp(cmd, "at ", 3)==0)
		{
			char *sp=strchr(cmd+3, ' ');
			if(sp)
			{
				at_set=true;
				if(cmd[3]=='c')
				{
					sscanf(cmd+4, "%d", &at);
					at=onbuf->curline+at;
				}
				else if((cmd[3]=='s') && ((on_set?on:curbuffer)<0)) // if the buffer being acted upon is a script buffer, allow the s flag
				{
					sscanf(cmd+4, "%d", &at);
					at=onbuf->scrline+at;
					if((on_set?on:curbuffer)==-2) // if it's the current buffer
						at--; // then subtract one because one is added to scrline as soon as the script line is read
				}
				else
				{
					sscanf(cmd+3, "%d", &at);
				}
				char *ca=strdup(cmd);
				strcpy(cmd, ca+(sp-cmd)+1);
				free(ca);
			}
		}
		// split cmd and args
		char * split = strchr(cmd, ' ');
		if(split==NULL)
		{
			split=cmd+strlen(cmd);
			split[1]=0;
		}
		*split++=0;
		while((strlen(cmd)!=0) && (cmd[strlen(cmd)-1]==' '))
			cmd[strlen(cmd)-1]=0;
		if(strcmp(cmd, "goto")==0)
		{
			if(scriptmode)
			{
				printf("goto: ");
				sbuf[0]=cmd_move(split, sbuf[0], true);
				if(sbuf[0].err!=0)
				{
					printf("Error %u\n", sbuf[0].err);
					cont=false;
				}
				else
				{
					if(more)
					{
						free(more);
						more=NULL;
					}
				}
			}
			else
			{
				printf("Not a script!\n");
			}
			match=true;
		}
		if(strcmp(cmd, "onerror")==0)
		{
			if(scriptmode)
			{
				char c=*split;
				if((c=='-') || (c=='+') || (c=='='))
				{
					bool neg=false;
					switch(c)
					{
						case '-':
							neg=true;
						case '+': // fallthrough
							sbuf[0].oeabs=false;
						break;
						case '=':
							sbuf[0].oeabs=true;
						break;
					}
					sscanf(split+1, "%d", &sbuf[0].onerror);
					if(neg)
						sbuf[0].onerror = -sbuf[0].onerror;
				}
				else
				{
					printf("onerror: bad flag %c\n", c);
				}
			}
			else
			{
				printf("Not a script!\n");
			}
			match=true;
		}
		if(strcmp(cmd, "help")==0)
		{
			cmd_help(split, nbuffers==0);
			match=true;
		}
		if(strcmp(cmd, "new")==0)
		{
			if(on_set)
			{
				(*onbuf)=cmd_close(split, (*onbuf), true);
				if((*onbuf).err==0)
				{
					(*onbuf)=cmd_new(split);
				}
				if(scriptmode && (*onbuf).err!=0)
				{
					cont=false;
				}
			}
			else
			{
				buffer * newbuf=(buffer *)realloc(ebuf, (nbuffers+1)*sizeof(buffer));
				if(newbuf==NULL)
				{
					printf("new: A strange error occurred\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					ebuf=newbuf;
					ebuf[nbuffers]=cmd_new(split);
					curbuffer=nbuffers;
					if(scriptmode && ebuf[nbuffers].err!=0)
						cont=false;
					nbuffers++;
				}
			}
			match=true;
		}
		if(strcmp(cmd, "load")==0)
		{
			if(on_set && (on<nbuffers))
			{
				(*onbuf)=cmd_close(split, (*onbuf), false);
				if((*onbuf).err==0)
					(*onbuf)=cmd_load(split, true);
				if(scriptmode && (*onbuf).err!=0)
					cont=false;
			}
			else
			{
				buffer * newbuf=(buffer *)realloc(ebuf, (nbuffers+1)*sizeof(buffer));
				if(newbuf==NULL)
				{
					printf("load: A strange error occurred\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					ebuf=newbuf;
					ebuf[nbuffers]=cmd_load(split, true);
					curbuffer=nbuffers;
					if(scriptmode && ebuf[nbuffers].err!=0)
						cont=false;
					nbuffers++;
				}
			}
			match=true;
		}
		if(strcmp(cmd, "script")==0)
		{
			buffer nsbuf=cmd_load(split, true);
			if(scriptmode && nsbuf.err!=0)
			{
				printf("script: Failed to load script\n");
				cont=false;
			}
			else
			{
				if(!scriptmode)
					userscript=true;
				scriptmode=true;
				buffer *asbuf=script_push(sbuf, &sbufs, nsbuf);
				if(asbuf)
				{
					sbuf=asbuf;
					if(more)
					{
						free(more);
						more=NULL;
					}
				}
				else
				{
					printf("script: Failed to call script\n");
					if(scriptmode)
						cont=false;
				}
			}
			match=true;
		}
		if(strcmp(cmd, "quit")==0)
		{
			if(scriptmode)
				cont=false;
			else
			{
				if((nbuffers<2) && ((ebuf[0].buf==NULL) || ebuf[0].saved==true))
					cont=false;
				else
				{
					if(strchr(split, '!'))
						cont=false;
					else
					{
						printf("quit: Unsaved changes, or multiple open buffers!  Use quit ! to quit anyway\n");
					}
				}
			}
			match=true;
		}
		/*if(nbuffers==0)
		{
			if(!match)
			{
				printf("No buffers open!\n");
				if(scriptmode)
					cont=false;
			}
		}
		else*/
		{
			if(strcmp(cmd, "buffer")==0)
			{
				/* 'buffer' command - switches between open buffers, or queries which buffer is open */
				if((strchr(split, '?')!=NULL) || (*split==0))
				{
					if((*cbuf).filename!=NULL)
					{
						printf("buffer: %s#%d of %u, '%s'\n", (*cbuf).saved?"":"*", curbuffer+1, nbuffers, (*cbuf).filename);
					}
					else
					{
						printf("buffer: %s#%d of %u, untitled\n", (*cbuf).saved?"":"*", curbuffer+1, nbuffers);
					}
				}
				else
				{
					/*if(nbuffers<2)
					{
						printf("No other open buffers!\n");
						if(scriptmode)
							cont=false;
					}
					else*/
					{
						int newbuffer=-3;
						sscanf(split, "%d", &newbuffer);
						newbuffer--;
						if((newbuffer>=-2) && (newbuffer<nbuffers))
						{
							curbuffer=newbuffer;
							if(curbuffer==-1)
							{
								cbuf=&abuf;
							}
							else if(curbuffer<=-2)
							{
								int ws=-2-curbuffer;
								if(ws<sbufs)
									cbuf=&sbuf[ws];
								else
									cbuf=&nbuf;
							}
							else if(curbuffer<nbuffers)
							{
								cbuf=&ebuf[curbuffer];
							}
							else
							{
								cbuf=&nbuf;
							}
							if((*cbuf).filename!=NULL)
							{
								printf("buffer: Selected #%d '%s'\n", curbuffer+1, (*cbuf).filename);
							}
							else
							{
								printf("buffer: Selected #%d (untitled)\n", curbuffer+1);
							}
							if(scriptmode && ebuf[curbuffer].err!=0)
								cont=false;
						}
						else
						{
							printf("buffer: number out of range!\n");
							if(scriptmode)
								cont=false;
						}
					}
				}
				match=true;
			}
			if(on_set)
				cbuf=onbuf;
			(*cbuf).cline=(*cbuf).curline;
			if(at_set)
				(*cbuf).curline=max(min(at, (*cbuf).nlines-1), 0);
			if(strcmp(cmd, "save")==0)
			{
				if((*cbuf).buf==NULL)
				{
					printf("save: No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					(*cbuf)=cmd_save(split, (*cbuf));
					if(scriptmode && (*cbuf).err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "showall")==0)
			{
				if((*cbuf).buf==NULL)
				{
					printf("showall: No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					int e=cmd_show("-1,-1", (*cbuf));
					if(scriptmode && e!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "showpg")==0)
			{
				if((*cbuf).buf==NULL)
				{
					printf("showpg: No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					int e=cmd_showpg((*cbuf));
					if(scriptmode && e!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "show")==0)
			{
				if((*cbuf).buf==NULL)
				{
					printf("show: No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					int e=cmd_show(split, (*cbuf));
					if(scriptmode && e!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "move")==0)
			{
				if((*cbuf).buf==NULL)
				{
					printf("move: No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					(*cbuf)=cmd_move(split, (*cbuf), false);
					if(scriptmode && (*cbuf).err!=0)
						cont=false;
				}
				at_set=false;
				match=true;
			}
			if(strcmp(cmd, "insert")==0)
			{
				if((*cbuf).buf==NULL)
				{
					printf("insert: No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					(*cbuf)=cmd_insert(split, (*cbuf));
					if(scriptmode && (*cbuf).err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "copy")==0)
			{
				if((*cbuf).buf==NULL)
				{
					printf("copy: No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					int e=cmd_copy(split, (*cbuf));
					if(scriptmode && e!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "paste")==0)
			{
				if((*cbuf).buf==NULL)
				{
					printf("paste: No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					(*cbuf)=cmd_paste(split, (*cbuf));
					if(scriptmode && (*cbuf).err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "replace")==0)
			{
				if((*cbuf).buf==NULL)
				{
					printf("replace: No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					(*cbuf)=cmd_replace(split, (*cbuf));
					if(scriptmode && (*cbuf).err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "split")==0)
			{
				if((*cbuf).buf==NULL)
				{
					printf("split: No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					(*cbuf)=cmd_split(split, (*cbuf));
					if(scriptmode && (*cbuf).err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "append")==0)
			{
				if((*cbuf).buf==NULL)
				{
					printf("append: No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					(*cbuf)=cmd_append(split, (*cbuf));
					if(scriptmode && (*cbuf).err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "delete")==0)
			{
				if((*cbuf).buf==NULL)
				{
					printf("delete: No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					(*cbuf)=cmd_delete(split, (*cbuf));
					if(scriptmode && (*cbuf).err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "join")==0)
			{
				if((*cbuf).buf==NULL)
				{
					printf("join: No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					(*cbuf)=cmd_join(split, (*cbuf));
					if(scriptmode && (*cbuf).err!=0)
						cont=false;
				}
				match=true;
			}
			if(strcmp(cmd, "close")==0)
			{
				if((*cbuf).buf==NULL)
				{
					printf("close: No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else if((on_set?on:curbuffer)==-1)
				{
					printf("close: You can't close the alias-buffer!\n");
					if(scriptmode)
						cont=false;
				}
				else if((on_set?on:curbuffer)<=-2)
				{
					if((on_set?on:curbuffer)==-2)
						stop=true;
					else
						printf("close: No support for deep stack closes\n");
				}
				else
				{
					(*cbuf)=cmd_close(split, (*cbuf), true);
					if((*cbuf).buf==NULL)
					{
						nbuffers--;
						int i;
						for(i=on_set?on:curbuffer;i<nbuffers;i++)
						{
							ebuf[i]=ebuf[i+1];
						}
						if(curbuffer>=nbuffers)
							curbuffer=max(nbuffers-1, 0);
						if(nbuffers!=0)
						{
							buffer * newbuf=(buffer *)realloc(ebuf, nbuffers*sizeof(buffer));
							if(newbuf==NULL)
							{
								printf("close: A strange (non-critical) error occurred\n");
							}
							else
							{
								ebuf=newbuf;
							}
						}
					}
				}
				match=true;
			}
			if(strcmp(cmd, "count")==0)
			{
				if((*cbuf).buf==NULL)
				{
					printf("count: No file open!\n");
					if(scriptmode)
						cont=false;
				}
				else
				{
					int e=cmd_count(split, (*cbuf));
					if(scriptmode && e!=0)
						cont=false;
				}
				match=true;
			}
			if(at_set)
				(*cbuf).curline=(*cbuf).cline;
			if(!match)
			{
				printf("Command %s not recognised\n", cmd);
				if(scriptmode)
					cont=false;
			}
		}
		free(cmd);
		screrr:
		if((stop || !cont) && userscript && scriptmode)
		{
			if(!stop && ((sbuf[0].onerror>=0) || (sbuf[0].oeabs=false)))
			{
				if(sbuf[0].oeabs)
				{
					sbuf[0].scrline=sbuf[0].onerror;
				}
				else
				{
					sbuf[0].scrline+=sbuf[0].onerror;
				}
				sbuf[0].scrline=max(min(sbuf[0].scrline, sbuf[0].nlines-1), 0);
				cont=true;
				if(more)
				{
					free(more); // clear the more
					more=NULL;
				}
			}
			else
			{
				buffer *asbuf=script_pop(sbuf, &sbufs);
				if(asbuf)
				{
					sbuf=asbuf;
				}
				cont=true;
				if(more)
				{
					free(more); // clear the more
					more=NULL;
				}
			}
		}
	}
	
	return(0);
}

char * getl(char * prompt)
{
	printf("%s", prompt);
	fflush(stdout);
	// gets a line of string data, {re}alloc()ing as it goes, so you don't need to make a buffer for it, nor must thee fret thyself about overruns!
	char * lout = (char *)malloc(80);
	int i=0;
	char c;
	while(1)
	{
		c = getchar();
		if (c == 10)
			break;
		if (c != 0)
			lout[i++]=c;
		if ((i%80) == 0)
		{
			if ((lout = (char *)realloc(lout, i+80))==NULL)
			{
				free(lout);
				printf("\nNot enough memory to store input!\n");
				return(NULL);
			}
		}
	}
	lout[i]=0;
	lout=(char *)realloc(lout, i+1);
	return(lout);
}

char * slurp(FILE *fp)
{
	fseek(fp, 0, SEEK_SET);
	// slurps a filefull of textual data, {re}alloc()ing as it goes, so you don't need to make a buffer for it, nor must thee fret thyself about overruns!
	char * lout = (char *)malloc(80);
	int i=0;
	int c;
	while(!feof(fp))
	{
		c = fgetc(fp);
		if(c == EOF)
			break;
		if (c != 0)
			lout[i++]=(char)c;
		if ((i%80) == 0)
		{
			if ((lout = (char *)realloc(lout, i+80))==NULL)
			{
				free(lout);
				fprintf(stderr, "\nNot enough memory to store input!\n");
				return(NULL);
			}
		}
	}
	lout[i]=0;
	lout=(char *)realloc(lout, i+1);
	return(lout);
}

buffer cmd_new(char * split)
{
	/* 'new' command, creates a new buffer */
	buffer rv;
	rv.backup_depth=def_backup_depth;
	rv.nlines=0;
	rv.buf=(char **)malloc(rv.nlines*sizeof(char *));
	if(rv.buf==NULL)
	{
		printf("new: A strange error occurred\n");
		rv.err=1;
		return(rv);
	}
	rv.filename=strdup(split);
	rv.saved=true;
	rv.curline=0;
	rv.scrline=0;
	rv.err=0;
	return(rv);
}

buffer cmd_load(char * split, bool echo)
{
	/* 'load' command, loads a file from disk into a new buffer */
	buffer ebuf;
	ebuf.backup_depth=def_backup_depth;
	FILE *fp=fopen(split, "r");
	if(fp==NULL)
	{
		if(echo)
			printf("load: Failed to open file!\n");
		ebuf.err=1;
	}
	else
	{
		char * data = slurp(fp);
		fclose(fp);
		if(data==NULL)
		{
			if(echo)
				printf("Failed to load file!\n");
			ebuf.err=2;
		}
		else
		{
			ebuf.nlines=0;
			char * o=data;
			while(o!=NULL)
			{
				o=strchr(o+1, '\n');
				ebuf.nlines++;
			}
			ebuf.buf=(char **)malloc(ebuf.nlines*sizeof(char *));
			int line=0;
			o=data-1;
			while(o!=NULL)
			{
				char * next=strchr(o+1, '\n');
				if(next==NULL)
				{
					ebuf.buf[line]=strdup(o);
				}
				else
				{
					ebuf.buf[line]=(char *)malloc((size_t)(next-o));
					strncpy(ebuf.buf[line], o+1, next-o-1);
					ebuf.buf[line][next-o-1]=0;
					line++;
				}
				o=next;
			}
			ebuf.nlines--;
			ebuf.filename=strdup(split);
			if(echo)
				printf("load: Read %u bytes, %u lines\n", strlen(data), ebuf.nlines);
			free(data);
			ebuf.curline=0;
			ebuf.scrline=0;
			ebuf.err=0;
			ebuf.saved=true;
		}
	}
	return(ebuf);
}

buffer cmd_save(char * split, buffer ebuf)
{
	/* 'save' command, writes a buffer out to disk file */
	bool ignore_backup_fails=false;
	if(*split=='~')
	{
		ignore_backup_fails=true;
		split++;
	}
	if(strlen(split)>0)
		sscanf(split, "%s", ebuf.filename);
	if((ebuf.filename==NULL) || (*ebuf.filename==0))
	{
		printf("save: No filename specified!\n");
		ebuf.err=2;
		return(ebuf);
	}
	if(make_backups)
	{
		// check if file exists; if so, make a backup
		/* Note: This code is non-portable */
		if (access(ebuf.filename, F_OK) == 0)
		{
			int e=fbackup(ebuf.filename, 0, ebuf.backup_depth);
			if(e!=0)
			{
				printf("save: Failed to make backup!\n");
				if(ignore_backup_fails==false)
				{
					printf("File was not saved.\n");
					ebuf.err=3;
					return(ebuf);
				}
			}
		}
	}
	FILE *fp = fopen(ebuf.filename, "w");
	if(fp==NULL)
	{
		printf("save: Failed to open file '%s' for writing\n", ebuf.filename);
		ebuf.err=1;
	}
	else
	{
		int line;
		size_t bytes=0;
		for(line=0;line<ebuf.nlines;line++)
		{
			int p=0;
			while(ebuf.buf[line][p]!=0)
			{
				fputc(ebuf.buf[line][p], fp);
				bytes++;
				p++;
			}
			fputc('\n', fp); bytes++;
		}
		printf("save: Wrote %zu bytes, %u lines\n", bytes, ebuf.nlines);
		fclose(fp);
		ebuf.saved=true;
		ebuf.err=0;
	}
	return(ebuf);
}

int cmd_show(char * split, buffer ebuf)
{
	/* 'show' command, displays current line (and context) of buffer on-screen */
	int scrcols=80;
	char * envcols = getenv("COLUMNS");
	int showline=-1;
	int context=2;
	sscanf(split, "%d,%d", &showline, &context); // if there isn't one, it should leave it unchanged
	if(showline==-1)
		showline=ebuf.curline;
	if(context==-1)
		context=ebuf.nlines;
	if(envcols!=NULL)
		sscanf(envcols, "%u", &scrcols);
	if(scrcols<=0)
		scrcols=80;
	char * rule = ruler(scrcols);
	printf("%s\n", rule);
	int dline;
	for(dline=-context;dline<context+1;dline++)
	{
		int line=showline+dline;
		if((line>=0) && (line<ebuf.nlines))
		{
			printf("%05u%c", line, line==ebuf.curline?'#':line==ebuf.scrline?'@':':');
			int p;
			for(p=0;ebuf.buf[line][p]!=0;p++)
			{
				if(ebuf.buf[line][p]=='\t')
					printf("    ");
				else
					printf("%c", ebuf.buf[line][p]);
			}
			printf("$\n");
		}
	}
	if(context==ebuf.nlines)
		printf("%s\n", rule);
	return(0);
}

int cmd_showpg(buffer ebuf)
{
	/* 'showpg' command, shows the whole buffer, one page at a time */
	int line;
	int slines=0;
	int scrlines=24;
	int scrcols=80;
	char * envlines = getenv("LINES"), * envcols = getenv("COLUMNS");
	if(envlines!=NULL)
		sscanf(envlines, "%u", &scrlines);
	if(envcols!=NULL)
		sscanf(envcols, "%u", &scrcols);
	if(scrlines<=0)
		scrlines=24;
	if(scrcols<=0)
		scrcols=80;
	char * rule = ruler(scrcols);
	printf("%s\n", rule);
	bool cont=true;
	for(line=0;line<ebuf.nlines;line++)
	{
		printf("%05u%c", line, line==ebuf.curline?'#':line==ebuf.scrline?'@':':');
		int schars=5;
		int p;
		for(p=0;ebuf.buf[line][p]!=0;p++)
		{
			if(ebuf.buf[line][p]=='\t')
			{
				printf("    ");
				schars+=4;
			}
			else
			{
				printf("%c", ebuf.buf[line][p]);
				schars++;
			}
			if(schars>=scrcols)
			{
				schars-=scrcols;
				slines++;
			}
			if(slines+3>scrlines)
			{
				printf("\n%s", rule);
				fflush(stdout);
				char c;
				while(1)
				{
					c=getchar();
					if(c=='\n')
						break;
					if(c=='x')
						cont=false;
				}
				slines=0;
			}
			if(!cont)
				break;
		}
		if(cont)
			printf("$\n");
		slines++;
		if(slines+3>scrlines)
		{
			printf("%s", rule);
			fflush(stdout);
			char c;
			while(1)
			{
				c=getchar();
				if(c=='\n')
					break;
				if(c=='x')
					cont=false;
			}
			slines=0;
		}
		if(!cont)
			break;
	}
	free(rule);
	return(0);
}

buffer cmd_move(char * split, buffer ebuf, bool scrline)
{
	/* 'move' command, sets the current line */
	char movetype='=';
	int moveline=0;
	if((*split=='=') || (*split=='+') || (*split=='-') || (*split=='$'))
	{
		movetype=*(split++);
	}
	sscanf(split, "%u", &moveline);
	switch(movetype)
	{
		case '=':
		break;
		case '+':
			moveline+=scrline?ebuf.scrline-1:ebuf.curline;
		break;
		case '-':
			moveline=scrline?ebuf.scrline-moveline-1:ebuf.curline-moveline;
		break;
		case '$':
			moveline=ebuf.nlines-1;
		break;
	}
	if((moveline>=0) && (moveline<ebuf.nlines))
	{
		if(scrline)
			ebuf.scrline=moveline;
		else
			ebuf.curline=moveline;
		printf("%s to line %u\n", scrline?"Jumped":"Moved", scrline?ebuf.scrline:ebuf.curline);
		ebuf.err=0;
	}
	else
	{
		printf("%sLine number out of file\n", scrline?"":"move: ");
		ebuf.err=1;
	}
	return(ebuf);
}

buffer cmd_insert(char * split, buffer ebuf)
{
	/* 'insert' command, adds one or more new lines to the buffer */
	bool rpt=(strchr(split, '.')!=NULL);
	bool first=true;
	if(rpt)
		printf("To end, enter a dot (.) on a line by itself\n");
	char * line;
	insert:
	line=getl("(ins) ");
	if(line==NULL)
	{
		printf("insert: A strange error occurred\n");
		ebuf.err=1;
	}
	else
	{
		if(rpt && (strcmp(line, ".")==0))
		{
			printf("Line(s) inserted\n");
			free(line);
			ebuf.err=0;
		}
		else
		{
			char ** newbuf=(char **)realloc(ebuf.buf, (1+ebuf.nlines)*sizeof(char *));
			if(newbuf==NULL)
			{
				printf("insert: A strange error occurred\n");
				ebuf.err=1;
			}
			else
			{
				if(strchr(split,'+') && first && ebuf.curline<ebuf.nlines)
					ebuf.curline++;
				first=false;
				ebuf.buf=newbuf;
				int i;
				for(i=ebuf.nlines;i>ebuf.curline;i--)
				{
					ebuf.buf[i]=ebuf.buf[i-1];
				}
				ebuf.buf[ebuf.curline]=strdup(line);
				free(line);
				ebuf.nlines++;
				ebuf.saved=false;
				if(rpt)
				{
					ebuf.curline++;
					goto insert;
				}
				printf("Line inserted\n");
				ebuf.err=0;
			}
		}
	}
	return(ebuf);
}

buffer cmd_replace(char * split, buffer ebuf)
{
	/* 'replace' command, overwrites the current line (but stores the old line in the pastebuffer) */
	if(ebuf.nlines==0)
	{
		printf("replace: The current file is empty!\n");
		return(ebuf);
	}
	char * line=getl("(rep) ");
	if(line==NULL)
	{
		printf("replace: A strange error occurred\n");
		ebuf.err=1;
	}
	else
	{
		if(paste)
			free(paste);
		paste=ebuf.buf[ebuf.curline];
		ebuf.buf[ebuf.curline]=line;
		ebuf.saved=false;
		printf("Line replaced\n");
		ebuf.err=0;
	}
	return(ebuf);
}

buffer cmd_append(char * split, buffer ebuf)
{
	/* 'append' command, adds a line to the end of the buffer */
	bool rpt=false;
	if(strchr(split, '.')!=NULL)
	{
		rpt=true;
		printf("To end, enter a dot (.) on a line by itself\n");
	}
	char * line;
	append:
	line=getl("(app) ");
	if(line==NULL)
	{
		printf("append: A strange error occurred\n");
		ebuf.err=1;
	}
	else if(rpt && (strcmp(line, ".")==0))
	{
		printf("Line(s) appended\n");
		free(line);
		ebuf.err=0;
	}
	else
	{
		char ** newbuf=(char **)realloc(ebuf.buf, (1+ebuf.nlines)*sizeof(char *));
		if(newbuf==NULL)
		{
			printf("append: A strange error occurred\n");
			ebuf.err=1;
		}
		else
		{
			ebuf.buf=newbuf;
			ebuf.buf[ebuf.nlines]=line;
			ebuf.curline=ebuf.nlines;
			ebuf.nlines++;
			ebuf.saved=false;
			if(rpt)
				goto append;
			printf("Line appended\n");
			ebuf.err=0;
		}
	}
	return(ebuf);
}

buffer cmd_split(char * split, buffer ebuf)
{
	/* 'split' command, splits the current line into two */
	if(ebuf.nlines==0)
	{
		printf("split: The current file is empty!\n");
		return(ebuf);
	}
	int pos=0;
	sscanf(split, "%d", &pos);
	int p=0,q=0;
	if(pos>=0)
	{
		for(q=0;q<=strlen(ebuf.buf[ebuf.curline]);q++)
		{
			if(ebuf.buf[ebuf.curline][q]=='\t')
				p+=4;
			else
				p++;
			if(p>pos)
				break;
		}
	}
	else
	{
		p++;
		for(q=strlen(ebuf.buf[ebuf.curline])-1;q>=0;q--)
		{
			if(ebuf.buf[ebuf.curline][q]=='\t')
				p+=4;
			else
				p++;
			if(p>-pos)
				break;
		}
	}
	if(q>=strlen(ebuf.buf[ebuf.curline]))
	{
		printf("split: Column number out of line\n");
		ebuf.err=1;
	}
	else
	{
		char * line=strdup(ebuf.buf[ebuf.curline]+q);
		if(line==NULL)
		{
			printf("split: A strange error occurred\n");
			ebuf.err=2;
		}
		else
		{
			char ** newbuf=(char **)realloc(ebuf.buf, (1+ebuf.nlines)*sizeof(char *));
			if(newbuf==NULL)
			{
				printf("split: A strange error occurred\n");
				ebuf.err=2;
			}
			else
			{
				ebuf.buf[ebuf.curline][q]=0;
				ebuf.buf[ebuf.curline]=(char *)realloc(ebuf.buf[ebuf.curline], max(q, 1));
				ebuf.curline++;
				ebuf.buf=newbuf;
				int i;
				for(i=ebuf.nlines;i>ebuf.curline;i--)
				{
					ebuf.buf[i]=ebuf.buf[i-1];
				}
				ebuf.buf[ebuf.curline]=line;
				ebuf.nlines++;
				ebuf.saved=false;
				printf("Line split\n");
				ebuf.err=0;
			}
		}
	}
	return(ebuf);
}

buffer cmd_join(char * split, buffer ebuf)
{
	/* 'join' command, joins two lines together */
	if(ebuf.nlines<2)
	{
		printf("Not enough lines to join!\n");
		return(ebuf);
	}
	if((strchr(split, '+')==NULL) || ebuf.curline+1==ebuf.nlines)
	{
		if(ebuf.curline>0)
			ebuf.curline--;
	}
	ebuf.buf[ebuf.curline]=(char *)realloc(ebuf.buf[ebuf.curline], strlen(ebuf.buf[ebuf.curline])+strlen(ebuf.buf[ebuf.curline+1])+1);
	strcat(ebuf.buf[ebuf.curline], ebuf.buf[ebuf.curline+1]);
	free(ebuf.buf[ebuf.curline+1]);
	int i;
	ebuf.nlines--;
	for(i=ebuf.curline+1;i<ebuf.nlines;i++)
	{
		ebuf.buf[i]=ebuf.buf[i+1];
	}
	ebuf.saved=false;
	if(ebuf.nlines>0)
	{
		char ** newbuf=(char **)realloc(ebuf.buf, (ebuf.nlines)*sizeof(char *));
		if(newbuf==NULL)
			printf("join: A strange error occurred\n");
		else
			ebuf.buf=newbuf;
	}
	printf("Lines joined\n");
	ebuf.err=0;
	return(ebuf);
}

buffer cmd_delete(char * split, buffer ebuf)
{
	/* 'delete' command, removes a line from the buffer (but stores it in the pastebuffer) */
	if(ebuf.nlines==0)
	{
		printf("delete: The current file is empty!\n");
		return(ebuf);
	}
	if(paste)
		free(paste);
	paste=ebuf.buf[ebuf.curline];
	int i;
	ebuf.nlines--;
	for(i=ebuf.curline;i<ebuf.nlines;i++)
	{
		ebuf.buf[i]=ebuf.buf[i+1];
	}
	ebuf.saved=false;
	if(ebuf.nlines>0)
	{
		char ** newbuf=(char **)realloc(ebuf.buf, (ebuf.nlines)*sizeof(char *));
		if(newbuf==NULL)
			printf("delete: A strange (but non-critical) error occurred\n");
		else
			ebuf.buf=newbuf;
	}
	printf("Line deleted\n");
	ebuf.err=0;
	return(ebuf);
}

int cmd_copy(char * split, buffer ebuf)
{
	/* 'copy' command, copies the current line into the pastebuffer */
	if(ebuf.nlines==0)
	{
		printf("copy: The current file is empty!\n");
		return(1);
	}
	if(paste)
		free(paste);
	paste=strdup(ebuf.buf[ebuf.curline]);
	return(0);
}

buffer cmd_paste(char * split, buffer ebuf)
{
	/* 'paste' command, inserts the contents of the pastebuffer at the current line */
	if(paste==NULL)
	{
		printf("There is nothing to paste\n");
		ebuf.err=2;
	}
	else
	{
		char ** newbuf=(char **)realloc(ebuf.buf, (1+ebuf.nlines)*sizeof(char *));
		if(newbuf==NULL)
		{
			printf("paste: A strange error occurred\n");
			ebuf.err=1;
		}
		else
		{
			if(strchr(split,'+') && ebuf.nlines!=0)
				ebuf.curline++;
			ebuf.buf=newbuf;
			int i;
			for(i=ebuf.nlines;i>ebuf.curline;i--)
			{
				ebuf.buf[i]=ebuf.buf[i-1];
			}
			ebuf.buf[ebuf.curline]=strdup(paste);
			ebuf.nlines++;
			ebuf.saved=false;
			printf("Line pasted\n");
			ebuf.err=0;
		}
	}
	return(ebuf);
}

int cmd_count(char * split, buffer ebuf)
{
	/* 'count' command, counts bytes, words and lines */
	int words=0;
	size_t bytes=0;
	int line;
	for(line=0;line<ebuf.nlines;line++)
	{
		bytes+=strlen(ebuf.buf[line])+1;
		char * o=strdup(ebuf.buf[line]);
		char * p=strtok(o, " \t");
		while(p!=NULL)
		{
			words++;
			p=strtok(NULL, " \t");
		}
		free(o);
	}
	printf("Count: %zu bytes, %u words, %u lines\n", bytes, words, ebuf.nlines);
	return(0);
}

buffer cmd_close(char * split, buffer ebuf, bool echo)
{
	/* 'close' command, closes a buffer */
	if(strchr(split, '!') || ebuf.saved)
	{
		if(ebuf.buf)
		{
			int i;
			for(i=0;i<ebuf.nlines;i++)
			{
				if(ebuf.buf[i])
					free(ebuf.buf[i]);
			}
			free(ebuf.buf);
		}
		ebuf.buf=NULL;
		if(ebuf.filename)
			free(ebuf.filename);
		ebuf.filename=NULL;
		if(echo)
			printf("Closed file\n");
		ebuf.err=0;
	}
	else
	{
		if(echo)
			printf("close: Unsaved changes!  Use close ! to close anyway\n");
		ebuf.err=1;
	}
	return(ebuf);
}

int cmd_help(char * split, bool nobufs)
{
	/* 'help' command, provides on-line documentation */
	int rv=0;
	if(*split!=0)
	{
		buffer hbuf=cmd_load(PREFIX"/share/doc/ted/cmdhelp", false);
		if(hbuf.err)
		{
			printf("help: It appears that you do not have ted command help installed.\n");
			rv=128+hbuf.err;
		}
		else
		{
			int i,j=0;
			for(i=0;i<hbuf.nlines;i++)
			{
				char * line=hbuf.buf[i];
				char * colon=strchr(line, ':');
				*colon++=0;
				if(strcmp(split, line)==0)
				{
					char * ptr=strchr(colon, '~');
					while(ptr!=NULL)
					{
						*ptr++='\n';
						ptr=strchr(ptr, '~');
					}
					printf("%s: %s\n", line, colon);
					j++;
				}
			}
			if(j==0)
			{
				printf("[help]: no such command '%s'\n", split);
				rv=128;
			}
		}
		cmd_close("", hbuf, false);
	}
	else
	{
		buffer hbuf=cmd_load(PREFIX"/share/doc/ted/readme", false);
		if(hbuf.err)
		{
			printf("help: It appears that you do not have ted help installed.\n");
			rv=128+hbuf.err;
		}
		else
		{
			cmd_showpg(hbuf);
			cmd_close("", hbuf, false);
		}
	}
	return(rv);
}

char * ruler(int cols)
{
	/* generates a ruler */
	if(cols<8)
		return(strdup(""));
	char * rv=(char *)malloc(cols);
	int off=0;
	off=sprintf(rv, "LINE: |");
	while(off+20<cols)
	{
		off+=sprintf(rv+off, "   |   | 0 |   |   0");
	}
	snprintf(rv+off, cols-off, "   |   | 0 |   |   0");
	return(rv);
}

int fbackup(char* fname, int rec, int backups)
{
	// backs up fname to a depth of backups.  rec tells us how deep we've gone (this function is recursive)
	// summary of how it works:
	//  it winds successive fbackup()s onto the stack for as many backups of the file as there are, then it backs up each file~n to file~(n+1)
	//  and in doing so, unwinds back down the stack, moving each file onto the space vacated by moving the previous one.
	//  if we hit the backups-th backup, we discard it, and unwind as usual.
	char * fnm = (char *)malloc(strlen(fname)+5+3*sizeof(int));
	FILE *exists_file;
	if ((exists_file = fopen(fname, "r")) != NULL)
	{
		// if the file exists, back it up
		fclose(exists_file);
		if (rec==0)
		{
			// when first called, we want to save somefile as somefile~1
			//  but first, we need to backup somefile~1 to somefile~2 if it exists - so we call fbackup with rec=1
			strcpy(fnm, fname);
			strcat(fnm, "~");
			if (fbackup(fnm, 1, backups) != 0)
				return(1);
			if (rename(fname, fnm) != 0)
				return(3);
			//if the original file existed, then we've now successfully backed up - so say so.
			printf("Backed up OK.\n");
		}
		else
		{
			// if we've already recursed partway, then we want to save somefile~n as somefile~(n+1)
			//  but first, we need to backup somefile~(n+1) as somefile~(n+2) if it exists - so we call fbackup with rec=n+1
			char *i;
			char buf[1+3*sizeof(int)];
			char srec[6] = "~";
			if(rec>1)
			{
				sprintf(buf, "%d", rec);
				strcat(srec, buf);
			}
			// we should be able to find ~n in the filename we've been passed (if not, we return a fail, and unwind down the stack)
			if ((i = strstr(fname, srec)) == NULL)
				return(5);
			// get the filename with the ~n chopped off
			int offset;
			offset = strlen(fname)-strlen(i);
			strncpy(fnm, fname, offset);
			fnm[offset]=0;
			rec++;
			// if we've reached backup number backups, then throw this one away and return success
			if (rec > backups)
			{
				printf("Discarded backup ~%d ...\n", rec);
				return 0;
			}
			// construct the ~(n+1) suffix to add to the end
			strcpy(srec, "~");
			if(rec>1)
			{
				sprintf(buf, "%d", rec);
				strcat(srec, buf);
			}
			strcat(fnm, srec);
			// backup somefile~(n+1) to somefile~(n+2) if it exists.
			if (fbackup(fnm, rec, backups) != 0)
				return(4);
			// rename somefile~n to somefile~(n+1)
			if (rename(fname, fnm) != 0)
				return(6);
			}
		}
	else
	{
		// if the file doesn't exist...
		if (rec != 0)
		{
			// ...but we've already recursed, then we must have run out of old backups to back up
			// so we say how deep we backed up, and then we return success, unwinding back down the stack
			// BUT... because we climb first, we haven't yet done the backups, so we say we're GOING to back up this far...
			//  so that if it crashes, the user doesn't assume it backed up fine.
			printf("Backing up to ~%d ...\n", rec);
		}
	}
	return(0);
}

buffer add_line_to_buffer(buffer ebuf, int line, char * data)
{
	/* adds a line to buffer <ebuf>, at the <line>-th line, and stores <data> in it */
	if(line==-1)
		line=ebuf.curline;
	char ** newbuf=(char **)realloc(ebuf.buf, (1+ebuf.nlines)*sizeof(char *));
	if(newbuf==NULL)
	{
		printf("A strange error occurred\n");
		ebuf.err=1;
	}
	else
	{
		ebuf.buf=newbuf;
		int i;
		for(i=ebuf.nlines;i>line;i--)
		{
			ebuf.buf[i]=ebuf.buf[i-1];
		}
		ebuf.buf[line]=data;
		ebuf.nlines++;
		ebuf.err=0;
	}
	return(ebuf);
}

buffer * script_pop(buffer *sbuf, int *sbufs)
{
	cmd_close("!", sbuf[0], false);
	int i;
	(*sbufs)--;
	for(i=0;i<*sbufs;i++)
	{
		sbuf[i]=sbuf[i+1];
	}
	buffer *asbuf = (buffer *)realloc(sbuf, max(*sbufs,1)*sizeof(buffer));
	if(asbuf)
		return(asbuf);
	else
	{
		printf("script_pop: A strange error occurred\n");
		return(NULL);
	}
}

buffer * script_push(buffer *sbuf, int *sbufs, buffer nsbuf)
{
	buffer *asbuf = (buffer *)realloc(sbuf, ((*sbufs)+1)*sizeof(buffer));
	if(asbuf)
	{
		int i;
		for(i=*sbufs;i>0;i--)
		{
			sbuf[i]=sbuf[i-1];
		}
		(*sbufs)++;
		sbuf[0]=nsbuf;
		return(asbuf);
	}
	else
	{
		printf("script: A strange error occurred\n");
		return(NULL);
	}
}
